/*
   Copyright (c) 2002-3, Andrew McNab, University of Manchester
   All rights reserved.

   Redistribution and use in source and binary forms, with or
   without modification, are permitted provided that the following
   conditions are met:

     o Redistributions of source code must retain the above
       copyright notice, this list of conditions and the following
       disclaimer. 
     o Redistributions in binary form must reproduce the above
       copyright notice, this list of conditions and the following
       disclaimer in the documentation and/or other materials
       provided with the distribution. 

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
   CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
   BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
   ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/

/*------------------------------------------------------------------------*
 * For more about GridSite: http://www.gridpp.ac.uk/gridsite/             *
 *------------------------------------------------------------------------*/

#ifndef VERSION
#define VERSION "0.0.0"
#endif

#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>  
#include <sys/time.h>            
#include <sys/wait.h>
#include <fcntl.h>
#include <pthread.h>
#include <ctype.h>
#include <curl/curl.h>

struct threaddata { char            *version;
                    int              stream;
                    int              verbose;
                    pthread_mutex_t *nextblockmutex; 
                    int             *nextblock;
                    int              myblock;
                    int              totalstreams;
                    int             *livestreams; 
                    char            *outfile;
                    char            *urlsuffix;
                    struct hostent  *hostaddrent;
                    int              addrs;
                    FILE            *fp;
                    struct timeval   starttime;
                    int              length; 
                    int              blocksize;
                    int              src_https;
                    char            *src_url;
                    char            *capath;
                    char            *keypath;
                    char            *certpath; } ;

size_t headlength(void *ptr, size_t size, size_t nmemb, void *p)
/* find the header with the length of this file and put in *p (pointing to
   an int in the calling function) */
{
  char      *s;
  size_t     realsize;
  
  realsize = size * nmemb;
  s = malloc(realsize + 1);
  memcpy(s, ptr, realsize);
  s[realsize] = '\0';
  sscanf(s, "Content-Length: %d", (int *) p);
 
  free(s);
  return realsize;
}

void onegetthread(struct threaddata *threaddata)
{
  int     transferred = 0, i;
  char    range[81], *errorbuf, *url;
  CURL   *curlhandle;
  long    usecstaken;
  struct  timeval endtime;

  curlhandle = curl_easy_init();
  
  threaddata->fp = fopen(threaddata->outfile, "r+");

  errorbuf = malloc(CURL_ERROR_SIZE);
  curl_easy_setopt(curlhandle, CURLOPT_ERRORBUFFER, errorbuf);
  curl_easy_setopt(curlhandle, CURLOPT_USERAGENT,   threaddata->version);
  curl_easy_setopt(curlhandle, CURLOPT_VERBOSE,     threaddata->verbose);

  i = threaddata->stream % threaddata->addrs;

  if (threaddata->src_https)
    {
      url = threaddata->src_url;
      curl_easy_setopt(curlhandle, CURLOPT_SSLENGINE,   NULL);
      curl_easy_setopt(curlhandle, CURLOPT_SSLCERTTYPE, "PEM");
      curl_easy_setopt(curlhandle, CURLOPT_SSLCERT,     threaddata->certpath);
      curl_easy_setopt(curlhandle, CURLOPT_SSLKEY,      threaddata->keypath);
      curl_easy_setopt(curlhandle, CURLOPT_CAPATH,      threaddata->capath);
    }
  else
    {
      url = malloc(24 + strlen(threaddata->urlsuffix)); 

      sprintf(url, "http://%d.%d.%d.%d%s", 
               (unsigned char) threaddata->hostaddrent->h_addr_list[i][0], 
               (unsigned char) threaddata->hostaddrent->h_addr_list[i][1], 
               (unsigned char) threaddata->hostaddrent->h_addr_list[i][2], 
               (unsigned char) threaddata->hostaddrent->h_addr_list[i][3], 
               threaddata->urlsuffix);
    }
 
  curl_easy_setopt(curlhandle, CURLOPT_URL,         url);
  curl_easy_setopt(curlhandle, CURLOPT_WRITEDATA,   threaddata->fp);

  while (1)
       {
         pthread_mutex_lock(threaddata->nextblockmutex);

         if (*(threaddata->nextblock) * threaddata->blocksize 
                                          >= threaddata->length)
           {    
             --(*(threaddata->livestreams));
             
             if (*(threaddata->livestreams) == 0) 
              /* we were last stream/thread: tidy up */
               {
                 fclose(threaddata->fp);
                 
                 curl_easy_cleanup(curlhandle);
                 
                 if (threaddata->verbose)
                  fprintf(stderr, "stream %d exits after %d bytes (last)\n", 
                          threaddata->stream, transferred);
                                
                 gettimeofday(&endtime, NULL);
                 
                 usecstaken = endtime.tv_sec * 1000000 +
                              endtime.tv_usec -
                             (threaddata->starttime.tv_sec * 1000000 +
                              threaddata->starttime.tv_usec);

                 if (threaddata->verbose)
                  fprintf(stderr, 
                         "%d bytes in %f seconds at %f KB/s with %d streams\n",
                         threaddata->length, usecstaken / 1000000.0,
                         threaddata->length / (usecstaken / 976.5625),
                         threaddata->totalstreams);

                 exit(0);
               }

             pthread_mutex_unlock(threaddata->nextblockmutex);
             curl_easy_cleanup(curlhandle);
           
             if (threaddata->verbose)
              fprintf(stderr, "stream %d exits after %d bytes\n", 
                      threaddata->stream, transferred);
                   
             return;
           }
  
         threaddata->myblock = *(threaddata->nextblock);
         ++(*(threaddata->nextblock));

         pthread_mutex_unlock(threaddata->nextblockmutex);

         if (threaddata->myblock >= threaddata->length / threaddata->blocksize)
           {
             sprintf(range, "%d-%d", 
                     threaddata->myblock * threaddata->blocksize, 
                     threaddata->length - 1);  
             transferred += threaddata->length - 
                                 threaddata->myblock * threaddata->blocksize;
           }
         else 
           {
             sprintf(range, "%d-%d",threaddata->myblock*threaddata->blocksize,
                        (threaddata->myblock + 1) * threaddata->blocksize - 1);
             transferred += threaddata->blocksize;
           }
  
         fseek(threaddata->fp, threaddata->blocksize *
                               threaddata->myblock, SEEK_SET);

         curl_easy_setopt(curlhandle, CURLOPT_RANGE, range);
         if (curl_easy_perform(curlhandle) != 0)
           {
             fprintf(stderr, "curl error: %s\n", errorbuf); 
             exit(5);
           }
      }
}

int main(int argn, char *argv[])
{
  int                 i, firstarg = 1, length = -1, blocksize = 1024 * 256, 
                      streams = 1, fd, nextblock = 0, livestreams, verbose = 0,
                      src_https = 0;
  char               *url = NULL, *outfile = NULL, *hostname, *p, *urlsuffix, 
                     *errorbuf, *version, *src_url;
  pthread_t           thread;
  pthread_mutex_t     nextblockmutex = PTHREAD_MUTEX_INITIALIZER;
  pthread_attr_t      threadattrs;
  struct hostent     *hostaddrent;
  CURL               *curlhandle;  
  
  char                capath[]   = "/etc/grid-security/certificates",
                     *certpath, *keypath;
  
  struct threaddata   threaddata, *tdatacpy; 

  for (i = 1; i < argn - 1; ++i)
    {
      if      ((strcmp(argv[i], "-s") == 0) && (argv[i+1][0] != '\0'))
        {
          streams = atoi(argv[i+1]);
          ++i;
        } 
      else if ((strcmp(argv[i], "-z") == 0) && (argv[i+1][0] != '\0'))
        {
          blocksize = atoi(argv[i+1]);
          ++i;
        } 
      else if (strcmp(argv[i], "-v") == 0) verbose = 1;
      else if (argv[i][0] != '-')
        {
          src_url = argv[i];
          outfile = argv[i+1];

          break;
        }
    }

  if ((src_url == NULL) ||
      (outfile == NULL) ||
      ((strncmp(src_url, "http://",  7) != 0) && 
       (strncmp(src_url, "https://", 8) != 0)))
    {
      fputs("Syntax: htcp [-s streams] [-z blocksize] [-v] URL file\n", stderr);
      return 1;
    }

  if (strncmp(src_url, "https", 5) == 0) 
    {
      src_https = 1;
      hostname = strdup(&src_url[8]);
      
      asprintf(&certpath, "/tmp/x509up_u%d", geteuid());
      asprintf(&keypath,  "/tmp/x509up_u%d", geteuid());
    }
  else hostname = strdup(&src_url[7]);
  
  p = index(hostname, ':');
  if (p == NULL) p = index(hostname, '/');
  if (p == NULL) p = "";
  
  urlsuffix = strdup(p);
  *p = '\0';

  if ((hostaddrent = gethostbyname(hostname)) == NULL)
    {
      fprintf(stderr, "Failed to look up hostname %s\n", hostname);
      return 2;
    }
  
  livestreams = streams;

  gettimeofday(&(threaddata.starttime), NULL);
    
  curlhandle = curl_easy_init();
  
  errorbuf = malloc(CURL_ERROR_SIZE);
  curl_easy_setopt(curlhandle, CURLOPT_ERRORBUFFER, errorbuf);

  version = malloc(sizeof("htcp version ") + sizeof(VERSION) + 1);
  strcpy(version, "htcp version ");
  strcat(version, VERSION);

  curl_easy_setopt(curlhandle, CURLOPT_USERAGENT, version);
  curl_easy_setopt(curlhandle, CURLOPT_NOBODY, 1);
  curl_easy_setopt(curlhandle, CURLOPT_HEADERFUNCTION, headlength);
  curl_easy_setopt(curlhandle, CURLOPT_WRITEHEADER, (void *) &length);
  curl_easy_setopt(curlhandle, CURLOPT_VERBOSE, verbose);

  if (src_https)
    {
      curl_easy_setopt(curlhandle, CURLOPT_SSLCERT,     certpath);
      curl_easy_setopt(curlhandle, CURLOPT_SSLKEY,      keypath);
      curl_easy_setopt(curlhandle, CURLOPT_CAPATH,      capath);
      curl_easy_setopt(curlhandle, CURLOPT_SSLENGINE,   NULL);
      curl_easy_setopt(curlhandle, CURLOPT_SSLCERTTYPE, "PEM");

      url = src_url;
    }
  else
    {    
      url = malloc(24 + strlen(urlsuffix)); 
      /* 24 = httpS://xxx.xxx.xxx.xxx + \0 */
  
      sprintf(url, "http://%d.%d.%d.%d%s", 
               (unsigned char) hostaddrent->h_addr_list[0][0], 
               (unsigned char) hostaddrent->h_addr_list[0][1], 
               (unsigned char) hostaddrent->h_addr_list[0][2], 
               (unsigned char) hostaddrent->h_addr_list[0][3], 
               urlsuffix);
    }

  curl_easy_setopt(curlhandle, CURLOPT_URL, url);

  if (curl_easy_perform(curlhandle) != 0)
    {
      fprintf(stderr, "curl error: %s\n", errorbuf); 
      return 2;
    }
  
  if (length < 0) 
    {
      fprintf(stderr, "Failed to get Content-Length for %s\n", url);
      return 3;
    }

  if ((fd = open(outfile, O_RDWR | O_CREAT | O_TRUNC, 0600)) == -1)
    {
      fprintf(stderr, "Failed to create output file %s\n", outfile);
      return 4;    
    }
  close(fd);

  pthread_attr_init(&threadattrs);
  pthread_attr_setdetachstate(&threadattrs, PTHREAD_CREATE_DETACHED);

  threaddata.verbose        = verbose;
  threaddata.version        = version;
  threaddata.nextblockmutex = &nextblockmutex;
  threaddata.nextblock      = &nextblock;
  threaddata.totalstreams   = streams;
  threaddata.livestreams    = &livestreams;
  threaddata.outfile        = outfile;
  threaddata.urlsuffix      = urlsuffix;
  threaddata.hostaddrent    = hostaddrent;  
  threaddata.length         = length;
  threaddata.blocksize      = blocksize;
  threaddata.src_https      = src_https;
  threaddata.src_url        = src_url;
  threaddata.capath         = capath;
  threaddata.certpath       = certpath;
  threaddata.keypath        = keypath;

  for (threaddata.addrs=0; 
       hostaddrent->h_addr_list[threaddata.addrs] != NULL;
       ++(threaddata.addrs)) ;

  for (i=0; i < streams; ++i)   
     {
       tdatacpy = malloc(sizeof(struct threaddata));
       memcpy(tdatacpy, &threaddata, sizeof(struct threaddata));

       tdatacpy->stream = i;

       pthread_create(&thread, &threadattrs, (void *) onegetthread, 
                                                (void *) tdatacpy);
     }

  pause(); /* clean up is done by last thread, but we must wait for it */
}
